"""Registry search views."""

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import UUID

from django.contrib.auth.mixins import LoginRequiredMixin, PermissionRequiredMixin
from django.http import JsonResponse
from django.utils.dateparse import parse_datetime
from django.views.generic import View
from grill import LogProperty, Properties

from ferda.backend import BACKEND, LOGGER
from ferda.constants import LOGGER_SESSION_KEY, LogEntryType, LoggerResult

from .mixins import RelatedContactsMixin


class AbstractSearchView(ABC, View):
    """Abstract search view class."""

    @abstractmethod
    def get_search_results(
        self,
        query_values: List[str],
        search_items: List[str],
        valid_from: Optional[datetime] = None,
        valid_to: Optional[datetime] = None,
    ) -> list:
        """Get search results."""

    @abstractmethod
    def get_object_info(self, object_id: UUID, history_id: Optional[UUID] = None) -> dict:
        """Get registry object info."""

    @abstractmethod
    def get_object_history(self, object_id: UUID) -> dict:
        """Get registry object history."""

    @abstractmethod
    def get_overview_items(self, object_info: dict, valid_at: Optional[datetime] = None) -> dict:
        """Get dictionary of overview object items (list differs depending on object type)."""

    @property
    @abstractmethod
    def object_history_ids_name(self) -> str:
        """Get object history ids identifier."""

    @property
    def search_history(self):
        """Return whether historic data should be searched."""
        return self.kwargs.get("search_history", False)

    @staticmethod
    def get_match_keys(matches: List[str]) -> Dict[str, List[str]]:
        """Get dictionary of match keys.

        Args:
            matches: List of dictionaries representing matches of registry objects.

        Returns:
            Dictionary.
            Keys are parts of the match item names before the first dot.
            Values are lists of all match items sharing the common part before the first dot.
        """
        match_keys: Dict[str, List[str]] = {}
        for full_key in matches:
            base_key = full_key.split(".", 1)[0]
            if base_key not in match_keys:
                match_keys[base_key] = [full_key]
            else:
                match_keys[base_key].append(full_key)
        return match_keys

    def get(self, request, *args, **kwargs):
        """Render context data as JSON."""
        context = self.get_context_data(**kwargs)
        return JsonResponse(context)

    def get_context_data(self, **kwargs) -> dict:
        """Get search view context."""
        context = {}

        query_values = self.request.GET.getlist("query")
        search_items = self.request.GET.getlist("items", None)
        valid_from = self.request.GET.get("valid_from")
        valid_to = self.request.GET.get("valid_to")

        if valid_from:
            valid_from = parse_datetime(valid_from)
        if valid_to:
            valid_to = parse_datetime(valid_to)

        search_results = []
        if query_values:
            matching_objects = self.get_search_results(
                query_values,
                search_items,
                valid_from=valid_from,
                valid_to=valid_to,
            )

            for match in matching_objects:
                result = match.copy()
                if self.search_history:
                    for idx, history in enumerate(match["histories"]):
                        object_info = self.get_object_info(
                            match["object_id"],
                            history[self.object_history_ids_name][-1],
                        )
                        self.update_search_match(result["histories"][idx], object_info)
                    object_history = self.get_object_history(match["object_id"])
                    last_known_info = self.get_object_info(
                        match["object_id"], object_history["timeline"][-1][self.object_history_id_name]
                    )
                    result.update(
                        self.get_overview_items(
                            last_known_info,
                            object_history.get("valid_to"),
                        )
                    )
                    result["events"] = last_known_info["events"]
                else:
                    object_info = self.get_object_info(match["object_id"])
                    self.update_search_match(result, object_info)
                search_results.append(result)

        context["search_results"] = search_results
        return context

    def update_search_match(self, match: Dict[str, Any], object_info: Dict[str, Any]):
        """Update match object with object info."""
        match_keys = self.get_match_keys(match["matched_items"])
        for key in match_keys:
            match[key] = object_info.get(key, None)

        # Main items are displayed in any case (whether they match or not)
        match.update(self.get_overview_items(object_info, match.get("valid_to")))

        match["match_keys"] = match_keys
        return match


class SearchContactView(LoginRequiredMixin, PermissionRequiredMixin, AbstractSearchView):
    """Search contact view."""

    permission_required = "ferda.can_view_registry"
    object_history_id_name = "contact_history_id"
    object_history_ids_name = "contact_history_ids"

    SEARCH_ITEMS_MAP = {
        "handle": ["contact_handle"],
        "name": ["name"],
        "organization": ["organization", "address.company"],
        "telephone": ["telephone", "fax"],
        "email": ["email", "notify_email"],
        "address": [
            "place.street",
            "place.city",
            "place.state_or_province",
            "place.postal_code",
            "address.street",
            "address.city",
            "address.state_or_province",
            "address.postal_code",
        ],
        "identifier": ["vat_identification_number", "additional_identifier"],
    }

    def get_object_info(self, object_id: UUID, history_id: Optional[UUID] = None) -> dict:
        """Get registry object info."""
        return BACKEND.get_contact_info(object_id, history_id)

    def get_object_history(self, object_id: UUID) -> dict:
        """Get contact history."""
        return BACKEND.get_contact_history(contact_id=object_id)

    def get_overview_items(self, object_info: dict, valid_at: Optional[datetime] = None) -> dict:
        """Get dictionary of overview contact items."""
        overview_items = {
            "contact_handle": object_info.get("contact_handle", None),
            "name": object_info.get("name", None),
            "organization": object_info.get("organization", None),
            "telephone": object_info.get("telephone", None),
            "email": object_info.get("email", None),
        }
        return overview_items

    def get_search_results(
        self,
        query_values: List[str],
        search_items: List[str],
        valid_from: Optional[datetime] = None,
        valid_to: Optional[datetime] = None,
    ) -> list:
        """Return matched contacts."""
        expanded_search_items: List[str] = []
        for item in search_items:
            expanded_search_items += self.SEARCH_ITEMS_MAP.get(item, [])

        log_properties: Properties = {
            "query_values": [LogProperty(query, None) for query in query_values],
            "search_item": [LogProperty(s, None) for s in expanded_search_items],
        }
        log_entry = LOGGER.create(
            LogEntryType.CONTACT_SEARCH_HISTORY if self.search_history else LogEntryType.CONTACT_SEARCH,
            source_ip=self.request.META.get("REMOTE_ADDR", ""),
            session_id=self.request.session[LOGGER_SESSION_KEY],
            properties=log_properties,
        )
        with log_entry:
            if self.search_history:
                matching_contacts = BACKEND.search_contact_history(
                    query_values=query_values,
                    limit=100,
                    searched_items=expanded_search_items,
                    start=valid_from,
                    end=valid_to,
                )
            else:
                matching_contacts = BACKEND.search_contact(
                    query_values=query_values, limit=100, searched_items=expanded_search_items
                )
            log_entry.references["contact"] = [str(i["object_id"]) for i in matching_contacts["results"]]
            log_entry.result = LoggerResult.SUCCESS
        return matching_contacts["results"]


class SearchDomainView(LoginRequiredMixin, PermissionRequiredMixin, AbstractSearchView):
    """Search domain view."""

    permission_required = "ferda.can_view_registry"
    object_history_id_name = "domain_history_id"
    object_history_ids_name = "domain_history_ids"

    def get_object_info(self, object_id: UUID, history_id: Optional[UUID] = None) -> dict:
        """Get registry object info."""
        return BACKEND.get_domain_info(object_id, history_id)

    def get_object_history(self, object_id: UUID) -> dict:
        """Get domain history."""
        return BACKEND.get_domain_history(domain_id=object_id)

    def get_overview_items(self, object_info: dict, valid_at: Optional[datetime] = None) -> dict:
        """Get dictionary of overview domain items."""
        overview_items = {
            "fqdn": object_info.get("fqdn", None),
            "expires_at": object_info.get("expires_at", None),
        }
        registrant = object_info.get("registrant", None)
        if registrant is not None:
            registrant_history_id: Optional[UUID] = None
            if valid_at:
                registrant_history = BACKEND.get_contact_history(contact_id=registrant, start=valid_at, end=valid_at)
                registrant_history_id = registrant_history["timeline"][0]["contact_history_id"]

            registrant_info = BACKEND.get_contact_info(registrant, registrant_history_id)
            overview_items["registrant"] = {
                "contact_id": str(registrant),
                "contact_handle": registrant_info["contact_handle"],
                "name": registrant_info["name"],
            }
        return overview_items

    def get_search_results(
        self,
        query_values: List[str],
        search_items: List[str],
        valid_from: Optional[datetime] = None,
        valid_to: Optional[datetime] = None,
    ) -> list:
        """Return matched domains."""
        SEARCH_ITEMS = ["fqdn"]

        # FIXME: This ugly hack is here bacause backend can't handle
        # more than one query value in domain search at the moment.
        # Once it's fixed, we can remove this workaround.
        query_values = [" ".join(query_values)]

        log_properties: Properties = {
            "query_values": [LogProperty(query, None) for query in query_values],
            "search_item": [LogProperty(s, None) for s in SEARCH_ITEMS],
        }
        log_entry = LOGGER.create(
            LogEntryType.DOMAIN_SEARCH_HISTORY if self.search_history else LogEntryType.DOMAIN_SEARCH,
            source_ip=self.request.META.get("REMOTE_ADDR", ""),
            session_id=self.request.session[LOGGER_SESSION_KEY],
            properties=log_properties,
        )
        with log_entry:
            if self.search_history:
                matching_domains = BACKEND.search_domain_history(
                    query_values=query_values,
                    limit=100,
                    searched_items=SEARCH_ITEMS,
                    start=valid_from,
                    end=valid_to,
                )
            else:
                matching_domains = BACKEND.search_domain(
                    query_values=query_values,
                    limit=100,
                    searched_items=SEARCH_ITEMS,
                )
            log_entry.references["domain"] = [str(i["object_id"]) for i in matching_domains["results"]]
            log_entry.result = LoggerResult.SUCCESS
        return matching_domains["results"]


class SearchNssetView(LoginRequiredMixin, PermissionRequiredMixin, RelatedContactsMixin, AbstractSearchView):
    """Search nsset view."""

    permission_required = "ferda.can_view_registry"
    object_history_id_name = "nsset_history_id"
    object_history_ids_name = "nsset_history_ids"

    def get_object_info(self, object_id: UUID, history_id: Optional[UUID] = None) -> dict:
        """Get registry object info."""
        return BACKEND.get_nsset_info(object_id, history_id)

    def get_object_history(self, object_id: UUID) -> dict:
        """Get nsset history."""
        return BACKEND.get_nsset_history(nsset_id=object_id)

    def get_overview_items(self, object_info: dict, valid_at: Optional[datetime] = None) -> dict:
        """Get dictionary of overview nsset items."""
        overview_items = {
            "nsset_handle": object_info.get("nsset_handle", None),
            "dns_hosts": object_info.get("dns_hosts", None),
            "technical_contacts": self.get_related_contacts_overview(
                object_info.get("technical_contacts", []),
                valid_at=valid_at,
            ),
        }
        return overview_items

    def get_search_results(
        self,
        query_values: List[str],
        search_items: List[str],
        valid_from: Optional[datetime] = None,
        valid_to: Optional[datetime] = None,
    ) -> list:
        """Return matched nssets."""
        log_properties: Properties = {
            "query_values": [LogProperty(query, None) for query in query_values],
            "search_item": [LogProperty(s, None) for s in search_items],
        }
        log_entry = LOGGER.create(
            LogEntryType.NSSET_SEARCH_HISTORY if self.search_history else LogEntryType.NSSET_SEARCH,
            source_ip=self.request.META.get("REMOTE_ADDR", ""),
            session_id=self.request.session[LOGGER_SESSION_KEY],
            properties=log_properties,
        )
        with log_entry:
            if self.search_history:
                matching_nssets = BACKEND.search_nsset_history(
                    query_values=query_values,
                    limit=100,
                    searched_items=search_items,
                    start=valid_from,
                    end=valid_to,
                )
            else:
                matching_nssets = BACKEND.search_nsset(
                    query_values=query_values,
                    limit=100,
                    searched_items=search_items,
                )
            log_entry.references["nsset"] = [str(i["object_id"]) for i in matching_nssets["results"]]
            log_entry.result = LoggerResult.SUCCESS
        return matching_nssets["results"]


class SearchKeysetView(LoginRequiredMixin, PermissionRequiredMixin, RelatedContactsMixin, AbstractSearchView):
    """Search keyset view."""

    permission_required = "ferda.can_view_registry"
    object_history_id_name = "keyset_history_id"
    object_history_ids_name = "keyset_history_ids"

    def get_object_info(self, object_id: UUID, history_id: Optional[UUID] = None) -> dict:
        """Get registry object info."""
        return BACKEND.get_keyset_info(object_id, history_id)

    def get_object_history(self, object_id: UUID) -> dict:
        """Get keyset history."""
        return BACKEND.get_keyset_history(keyset_id=object_id)

    def get_overview_items(self, object_info: dict, valid_at: Optional[datetime] = None) -> dict:
        """Get dictionary of overview keyset items."""
        overview_items = {
            "keyset_handle": object_info.get("keyset_handle", None),
            "dns_keys": object_info.get("dns_keys", None),
            "technical_contacts": self.get_related_contacts_overview(
                object_info.get("technical_contacts", []),
                valid_at=valid_at,
            ),
        }
        return overview_items

    def get_search_results(
        self,
        query_values: List[str],
        search_items: List[str],
        valid_from: Optional[datetime] = None,
        valid_to: Optional[datetime] = None,
    ) -> list:
        """Return matched keysets."""
        log_properties: Properties = {
            "query_values": [LogProperty(query, None) for query in query_values],
            "search_item": [LogProperty(s, None) for s in search_items],
        }
        log_entry = LOGGER.create(
            LogEntryType.KEYSET_SEARCH_HISTORY if self.search_history else LogEntryType.KEYSET_SEARCH,
            source_ip=self.request.META.get("REMOTE_ADDR", ""),
            session_id=self.request.session[LOGGER_SESSION_KEY],
            properties=log_properties,
        )
        with log_entry:
            if self.search_history:
                matching_keysets = BACKEND.search_keyset_history(
                    query_values=query_values,
                    limit=100,
                    searched_items=search_items,
                    start=valid_from,
                    end=valid_to,
                )
            else:
                matching_keysets = BACKEND.search_keyset(
                    query_values=query_values,
                    limit=100,
                    searched_items=search_items,
                )

            log_entry.references["keyset"] = [str(i["object_id"]) for i in matching_keysets["results"]]
            log_entry.result = LoggerResult.SUCCESS
        return matching_keysets["results"]
